import { tswCtx, isNotEmptySet } from './Ctx';
import { g_CurrentContext } from './Scope';
import { EventKind } from './types';
import { privates, tswElement } from './elm';
import { tswRawHtml, tswElementWithValueBase } from './htmlElements';
import { tswCtxAttr } from './CtxAttr';
import { tswCtxValue } from './CtxValue';
export class tswCtxNodeBase extends tswCtx {
    firstChild = null;
    lastChild = null;
    elementsWithRootEvents;
    refs;
    domChangeListeners;
    get dbg_firstChild() { return this.firstChild; }
    get dbg_lastChild() { return this.lastChild; }
    get dbg_elementsWithRootEvents() { return this.elementsWithRootEvents; }
    get dbg_domChangeListeners() { return this.domChangeListeners; }
    removeOldContent(parentNode) {
        this.notifySelfAndChildren((ctx, beforeChildren) => ctx.domChange(beforeChildren, false));
        this.removeChildren();
        if (this.firstChild)
            this.removeNodes(parentNode);
    }
    addNewContent(parentNode, nodeBefore, content) {
        const f = document.createDocumentFragment();
        this.addNodes(f, content);
        const ctxRoot = this.getRootCtx();
        ctxRoot.invokeBeforeAttach();
        parentNode.insertBefore(f, nodeBefore);
        this.notifySelfAndChildren((ctx, beforeChildren) => ctx.domChange(beforeChildren, true));
    }
    shouldBeAddedToParent() {
        return super.shouldBeAddedToParent() ||
            isNotEmptySet(this.refs) ||
            isNotEmptySet(this.elementsWithRootEvents) ||
            isNotEmptySet(this.domChangeListeners);
    }
    cleanup() {
        super.cleanup();
        this.detachEventHandlers();
        this.resetRefs();
        this.domChangeListeners = undefined;
    }
    addRef(ref) {
        if (this.refs == null)
            this.refs = new Set();
        this.refs.add(ref);
    }
    resetRefs() {
        if (this.refs) {
            this.refs.forEach(ref => {
                ref.set(null);
            });
            this.refs = undefined;
        }
    }
    detachEventHandlers() {
        if (this.elementsWithRootEvents) {
            this.elementsWithRootEvents.forEach(el => {
                const ctxRoot = this.getRootCtx();
                ctxRoot.detachElmEventHandlers(el);
            });
            this.elementsWithRootEvents.clear();
        }
    }
    domChange(beforeChildren, attach) {
        const domChangeListeners = this.domChangeListeners;
        if (domChangeListeners) {
            domChangeListeners.forEach(l => {
                const f = beforeChildren ?
                    (attach ? l.afterAttachPre : l.beforeDetachPre) :
                    (attach ? l.afterAttachPost : l.beforeDetachPost);
                if (f) {
                    f.call(l);
                }
                else {
                    const lold = l;
                    const f2 = beforeChildren && attach ? lold.afterAttach :
                        !beforeChildren && !attach ? lold.beforeDetach :
                            null;
                    if (f2) {
                        const oldName = attach ? 'afterAttach' : 'beforeDetach';
                        const newName = attach ? 'afterAttachPre' : 'beforeDetachPost';
                        console.warn(`Event handler '${oldName}' is obsolete. Please, use '${newName}' instead`);
                        f2.call(lold);
                    }
                }
            });
        }
    }
    replaceNode(nodeKind, oldNode, newNode) {
        let matched = false;
        switch (nodeKind) {
            case 0:
                if (this.firstChild == oldNode) {
                    this.firstChild = newNode;
                    matched = true;
                }
                break;
            case 1:
                if (this.lastChild == oldNode) {
                    this.lastChild = newNode;
                    matched = true;
                }
                break;
        }
        if (matched && this.ctxParent)
            this.ctxParent.replaceNode(nodeKind, oldNode, newNode);
    }
    removeNodes(parentNode) {
        let node = this.firstChild;
        while (true) {
            if (!node)
                throw new Error("node is null");
            const nextNode = node.nextSibling;
            parentNode.removeChild(node);
            if (node == this.lastChild)
                break;
            node = nextNode;
        }
    }
    addNodes(parentNode, content) {
        const parentLastChild = parentNode.lastChild;
        this.insertContent(parentNode, content);
        let firstAddedNode = parentLastChild == null ? parentNode.firstChild : parentLastChild.nextSibling;
        if (firstAddedNode == null) {
            const placeholderNode = document.createComment('placeholder');
            parentNode.appendChild(placeholderNode);
            firstAddedNode = placeholderNode;
        }
        this.firstChild = firstAddedNode;
        this.lastChild = parentNode.lastChild;
    }
    addEventHandlers(el, elmEventMapItem) {
        if (this.elementsWithRootEvents == null)
            this.elementsWithRootEvents = new Set();
        this.elementsWithRootEvents.add(el);
        const ctxRoot = this.getRootCtx();
        ctxRoot.attachElmEventHandler(el, elmEventMapItem);
    }
    insertContent(parentNode, item) {
        if (item == null || item === true || item === false || item === '')
            return;
        if (isDomChangeEventListener(item)) {
            if (this.domChangeListeners == null)
                this.domChangeListeners = new Set();
            this.domChangeListeners.add(item);
        }
        if (item instanceof Array) {
            for (const i of item) {
                this.insertContent(parentNode, i);
            }
        }
        else if (item instanceof Function || isPropDef(item) || isRenderer(item)) {
            const ctx = new tswCtxNodes(item);
            ctx.setup(this, parentNode);
        }
        else if (item instanceof tswRawHtml) {
            const el = document.createElement('div');
            el.innerHTML = item.value;
            while (el.firstChild)
                parentNode.appendChild(el.firstChild);
        }
        else if (item instanceof tswElement) {
            const children = item[privates.ElementGeneric.children]();
            const tagName = item[privates.ElementGeneric.tagName]();
            if (tagName) {
                const ns = item[privates.ElementGeneric.ns]();
                const el = ns ? document.createElementNS(ns, tagName) : document.createElement(tagName);
                this.createAttrs(el, item);
                this.insertContent(el, children);
                this.setupElement(el, item);
                parentNode.appendChild(el);
            }
            else {
                this.insertContent(parentNode, children);
            }
        }
        else {
            const s = item.toString();
            const n = document.createTextNode(s);
            parentNode.appendChild(n);
        }
    }
    setupElement(el, item) {
        const refs = item[privates.ElementGeneric.getRefs]();
        if (refs) {
            for (const ref of refs) {
                ref.set(el);
                this.addRef(ref);
            }
        }
        if (item instanceof tswElementWithValueBase) {
            const valInfos = item[privates.ElementWithValueBase.getValueInfos]();
            if (valInfos) {
                this.addValueContext(el, valInfos);
                this.addValueEventHandlers(el, valInfos);
            }
        }
        const events = item[privates.ElementGeneric.events]();
        if (events) {
            this.addEventListeners(events, el);
        }
    }
    addEventListeners(events, el) {
        for (const elmEventMapItem of events) {
            switch (elmEventMapItem.eventKind) {
                case EventKind.direct:
                    if (elmEventMapItem.eventType == 'dom') {
                        const preventDefault = elmEventMapItem.eventName.toLowerCase() == 'click' && el instanceof HTMLAnchorElement;
                        const handler = elmEventMapItem.handleEvent;
                        el.addEventListener(elmEventMapItem.eventName, function (e) {
                            if (preventDefault)
                                e.preventDefault();
                            handler.call(this, e, this);
                        });
                    }
                    break;
                case EventKind.onRoot:
                    {
                        this.addEventHandlers(el, elmEventMapItem);
                    }
                    break;
                default:
                    throw new Error(`Unsupported eventKind: ${elmEventMapItem.eventKind}`);
            }
        }
    }
    addValueContext(el, valInfos) {
        if (valInfos instanceof Array) {
            for (const valInfo of valInfos) {
                this.addValueContext(el, valInfo);
            }
        }
        else {
            const ctx = new tswCtxValue(el, valInfos.propName, valInfos.propVal);
            ctx.setup(this);
        }
    }
    addValueEventHandlers(el, valInfos) {
        el.addEventListener('input', inputHandler);
        el.addEventListener('change', inputHandler);
        function inputHandler(e) {
            handleValueEvent(valInfos);
            function handleValueEvent(valInfos) {
                if (valInfos instanceof Array) {
                    for (const valInfo of valInfos) {
                        handleValueEvent(valInfo);
                    }
                }
                else {
                    const value = el[valInfos.propName];
                    if (valInfos.propVal.set(value)) {
                        if (valInfos.onChange) {
                            valInfos.onChange(value);
                        }
                    }
                }
            }
        }
    }
    createAttrs(el, elm) {
        const elm_attrs = elm[privates.ElementGeneric.attrs]();
        if (elm_attrs) {
            const isAttrCaseSensitive = elm[privates.ElementGeneric.attrsCaseSensitive]();
            const attrs = this.createAttrMap(elm_attrs, isAttrCaseSensitive);
            attrs.forEach((attrValue, attrName) => {
                const ctx = new tswCtxAttr(el, attrName, attrValue);
                ctx.setup(this);
            });
        }
    }
    createAttrMap(elm_attrs, isAttrCaseSensitive) {
        const attrs = new Map();
        for (const a of elm_attrs) {
            if (a.attrName) {
                const attrName = isAttrCaseSensitive ? a.attrName : a.attrName.toLowerCase();
                const v = attrs.get(attrName);
                if (v == null) {
                    attrs.set(attrName, a.attrValue);
                }
                else if (v instanceof Array) {
                    const vals = v;
                    vals.push(a.attrValue);
                }
                else {
                    const vals = [];
                    vals.push(v);
                    vals.push(a.attrValue);
                    attrs.set(attrName, vals);
                }
            }
        }
        return attrs;
    }
}
function isPropDef(v) {
    return v != null && v.get instanceof Function;
}
function isRenderer(v) {
    return v != null && v.render instanceof Function;
}
function isDomChangeEventListener(v) {
    if (v == null)
        return false;
    const l = v;
    const lOld = v;
    return (l.afterAttachPost instanceof Function ||
        l.afterAttachPre instanceof Function ||
        l.beforeDetachPost instanceof Function ||
        l.beforeDetachPre instanceof Function ||
        lOld.afterAttach instanceof Function ||
        lOld.beforeDetach instanceof Function);
}
export class tswCtxNodes extends tswCtxNodeBase {
    content;
    ctxRoot;
    constructor(content) {
        super();
        this.content = content;
    }
    getRootCtx() {
        if (this.ctxRoot == null)
            throw new Error("this.ctxRoot == null");
        return this.ctxRoot;
    }
    setup(ctxParent, parentNode) {
        this.ctxRoot = ctxParent.getRootCtx();
        const content = this.createContent();
        this.addNodes(parentNode, content);
        this.addCtxToParent(ctxParent);
    }
    update() {
        if (this.ctxParent === null)
            return;
        if (!this.lastChild)
            throw new Error("this.lastChild is null");
        const parentNode = this.lastChild.parentNode;
        if (!parentNode)
            throw new Error("parentNode is null");
        const nodeBefore = this.lastChild.nextSibling;
        const firstChildOld = this.firstChild;
        const lastChildOld = this.lastChild;
        this.removeOldContent(parentNode);
        const content = this.createContent();
        this.addNewContent(parentNode, nodeBefore, content);
        if (this.ctxParent != null) {
            this.ctxParent.replaceNode(0, firstChildOld, this.firstChild);
            this.ctxParent.replaceNode(1, lastChildOld, this.lastChild);
        }
    }
    createContent() {
        return g_CurrentContext.use(this, () => this.getContent());
    }
    getContent() {
        if (this.content instanceof Function)
            return this.content();
        if (isPropDef(this.content))
            return this.content.get();
        if (isRenderer(this.content))
            return this.content.render();
        throw new Error("Not supported type of content");
    }
    get dbg_content() { return this.content; }
}
